# -----------------------------------------------------------------
# Copyright (c) 2015, AIT Austrian Institute of Technology GmbH.
# All rights reserved. See file FMIPP_LICENSE.txt for details.
# -----------------------------------------------------------------

'''This module provides a Python interface for the FMI++ Library.'''

import os.path, sys

# Retrieve current path.
install_dir = os.path.dirname( os.path.abspath( __file__ ) )

# Add directory containing the SWIG modules and external libraries.
sys.path.append( os.path.join( install_dir, "lib" ) ) 

# Import FMI++ wrapper.
from .fmippim import *

# Import utility scripts.
from .extractFMU import *
from .pathToURI import *
from .simplifyModelDescription import *

try:
	from .numeric import *
except Exception as ex:
	print( ex )

# Extra info regarding licenses.
def licenseInfo():
	import textwrap

	info1 = 'The FMI++ Python Interface is based on code from the FMI++ Library and BOOST. Also, it includes compiled libraries implementing the SUNDIALS CVODE integrator.'
	info2 = 'For detailed information on the respective licenses please refer to the license files provided here:'

	width = 65
	full_info = '\n{}\n\n{}\n\n{}\n'.format( textwrap.fill( info1, width ), textwrap.fill( info2, width ), os.path.join( install_dir, 'licenses' ) )
	
	print( full_info )



# -----------------------------------------------------------------
# Copyright (c) 2022, AIT Austrian Institute of Technology GmbH.
# All rights reserved. See file FMIPP_LICENSE.txt for details.
# -----------------------------------------------------------------

# Extract an FMU.
def extractFMU( fmuFilePath, outputDirPath, command = None ):
	'''Extract an FMU to a folder.

	fmuFilePath -- path to the FMU file (string)
	outputDirPath -- folder to which the FMU should be extracted (string)
	command (optional) -- specify the command to unzip the FMU (string)

	The command for unzipping should be given as a string, using tags '{fmu}'
	and '{dir}' as placeholders for the FMU file path and the output directory.
	For instance:
	  - unzip: 'unzip {fmu} -d {dir}'
	  - 7-zip: '"C:\\Program Files\\7-Zip\\7z.exe" -o{dir} x {fmu}'
	'''
	import os, zipfile

	try:
		import urllib.parse as urlparse, urllib.request as urllib
	except:
		pass

	try:
		import urlparse, urllib
	except:
		pass

	# Check if specified file is indeed a zip file.
	if not zipfile.is_zipfile( fmuFilePath ):
		print( '%s is not a valid ZIP archive' % fmuFilePath )
		return

	# Check if output directory exists.
	if not os.path.isdir( outputDirPath ):
		print( '%s is not a valid path' % outputDirPath )
		return

	try:
		# Extract FMU file name from complete path.
		fmuFileName = os.path.split( fmuFilePath )[1];
		# Split FMU file name (name & extension).
		fmuSplitFileName = os.path.splitext( fmuFileName )

		# Check if this is a valid FMU file name.
		if '.fmu' != fmuSplitFileName[1]:
			print( '%s is not a valid FMU file name' % fmuFileName )

		# Extract model name from FMU file name.
		fmuModelName = fmuSplitFileName[0]

		# Create sub-directory in output directory.
		extractDirPath = os.path.join( outputDirPath, fmuModelName )
		try:
			os.mkdir( extractDirPath )
		except OSError: # Directory already exists
			print( 'directory already exists: %s' % extractDirPath )


		if command is None:
			# Access FMU.
			fmu = zipfile.ZipFile( fmuFilePath, 'r' )

			# Extract FMU to output directory.
			fmu.extractall( extractDirPath )
		else:
			os.system( command.format( fmu = fmuFilePath, dir = extractDirPath ) )

		# Return URI to extracted FMU.
		return urlparse.urljoin( 'file:', urllib.pathname2url( extractDirPath ) )
	except:
		print( 'failed to extract file: %s' % fmuFilePath )


if __name__ == '__main__':

	import sys

	if len( sys.argv ) != 3:
		print( 'Usage:\n\tpython extractFMU.py <path-to-fmu> <path-to-output-dir>\n' )
		sys.exit()

	fmuFilePath = sys.argv[1]
	outputDirPath = sys.argv[2]

	extractDirURI = extractFMU( fmuFilePath, outputDirPath )
	print( 'extracted FMU to: %s' % extractDirURI )



# -----------------------------------------------------------------
# Copyright (c) 2022, AIT Austrian Institute of Technology GmbH.
# All rights reserved. See file FMIPP_LICENSE.txt for details.
# -----------------------------------------------------------------

'''This module provides a Python interface for the FMI++ Library.'''

import os.path, sys

# Retrieve current path.
install_dir = os.path.dirname( os.path.abspath( __file__ ) )

# Add directory containing the SWIG modules and external libraries.
sys.path.append( os.path.join( install_dir ) )

# Import FMI++ wrapper.
from .fmippex import *



# -----------------------------------------------------------------
# Copyright (c) 2022, AIT Austrian Institute of Technology GmbH.
# All rights reserved. See file FMIPP_LICENSE.txt for details.
# -----------------------------------------------------------------

#
# Collection of helper functions for creating FMU CS according to FMI 1.0
#
import platform

# Get templates for the XML model description depending on the FMI version.
def fmi1GetModelDescriptionTemplates( verbose, modules ):
    # Template string for XML model description.
    main = '<?xml version="1.0" encoding="UTF-8"?>\n<fmiModelDescription fmiVersion="1.0" modelName="__MODEL_NAME__" modelIdentifier="__MODEL_IDENTIFIER__" description="TRNSYS FMI CS export" generationTool="FMI++ TRNSYS Export Utility" generationDateAndTime="__DATE_AND_TIME__" variableNamingConvention="flat" numberOfContinuousStates="0" numberOfEventIndicators="0" author="__USER__" guid="{__GUID__}">\n\t<VendorAnnotations>\n\t\t<Tool name="python">\n\t\t\t<Executable arguments="run_backend___GUID__.py" executableURI="__PYTHON_URI__"/>\n\t\t</Tool>\n\t</VendorAnnotations>\n\t<ModelVariables>\n__MODEL_VARIABLES__\t</ModelVariables>\n\t<Implementation>\n\t\t<CoSimulation_Tool>\n\t\t\t<Capabilities canHandleVariableCommunicationStepSize="true" canHandleEvents="true" canRejectSteps="false" canInterpolateInputs="false" maxOutputDerivativeOrder="0" canRunAsynchronuously="false" canBeInstantiatedOnlyOncePerProcess="false" canNotUseMemoryManagementFunctions="true"/>\n\t\t\t<Model entryPoint="fmu://resources/" manualStart="false" type="application/x-python">__ADDITIONAL_FILES__</Model>\n\t\t</CoSimulation_Tool>\n\t</Implementation>\n</fmiModelDescription>'

    # Template string for XML model description of scalar variables.
    scalar_variable_node = '\t\t<ScalarVariable name="__VAR_NAME__" valueReference="__VAL_REF__" variability="__VARIABILITY__" causality="__CAUSALITY__">\n\t\t\t<__VAR_TYPE____START_VALUE__/>\n\t\t</ScalarVariable>\n'

    # Template string for XML model description of scalar unknown (not defined in FMI 1.0).
    scalar_unknown_node = ''

    return ( main, scalar_variable_node, scalar_unknown_node )


def fmi1addVariabilityAndCausalityToModelDescription( scalar_variable_description, type, is_input, is_parameter, verbose, modules ):
    if ( True == is_parameter ):
        scalar_variable_description = scalar_variable_description.replace( '__CAUSALITY__', 'input' )
        scalar_variable_description = scalar_variable_description.replace( '__VARIABILITY__', 'parameter' )
    elif ( True == is_input and False == is_parameter ):
        scalar_variable_description = scalar_variable_description.replace( '__CAUSALITY__', 'input' )
        if ( 'Real' == type ):
            scalar_variable_description = scalar_variable_description.replace( '__VARIABILITY__', 'continuous' )
        else:
            scalar_variable_description = scalar_variable_description.replace( '__VARIABILITY__', 'discrete' )
    elif ( False == is_input and False == is_parameter ):
        scalar_variable_description = scalar_variable_description.replace( '__CAUSALITY__', 'output' )
        if ( 'Real' == type ):
            scalar_variable_description = scalar_variable_description.replace( '__VARIABILITY__', 'continuous' )
        else:
            scalar_variable_description = scalar_variable_description.replace( '__VARIABILITY__', 'discrete' )
    return scalar_variable_description


# Add optional files to XML model description.
def fmi1AddOptionalFilesToModelDescription( optional_files, model_description, verbose, modules ):
    if ( 0 == len( optional_files ) ):
        model_description = model_description.replace( '__ADDITIONAL_FILES__', '' )
    else:
        additional_files_description = ''
        indent = '\n\t\t\t'

        for file_name in optional_files:
            additional_files_description += indent + '\t<File file=\"fmu://resources/' + modules.os.path.basename( file_name ) + '\"/>'
            if ( True == verbose ): modules.log( '[DEBUG] Added additional file to model description: ', modules.os.path.basename( file_name ) )
        additional_files_description += indent

        model_description = model_description.replace( '__ADDITIONAL_FILES__', additional_files_description )

    return model_description


# Create shared library for FMU.
def fmi1CreateSharedLibrary( fmi_model_identifier, verbose, modules ):
    # Define name and extension of shared library file based on platform.
    system = modules.platform.system()
    if system == 'Linux':
      fmu_shared_library_name = fmi_model_identifier + '.so'
    elif system == 'Windows':
      fmu_shared_library_name = fmi_model_identifier + '.dll'
    else: # system == 'Darwin'
      fmu_shared_library_name = fmi_model_identifier + '.dylib'

    # Check if batch file for build process exists.
    build_process_batch_file = modules.os.path.join( modules.os.path.dirname( __file__ ), 'scripts', 'fmi1_build.bat' )
    if ( False == modules.os.path.isfile( build_process_batch_file ) ):
        raise RuntimeError( '\n[ERROR] Could not find file: {}'.format( build_process_batch_file ) )
    # Remove possible leftovers from previous builds.
    for file_name in modules.glob.glob( fmi_model_identifier + '.*' ): modules.os.remove( file_name )
    if ( True == modules.os.path.isfile( 'fmiFunctions.obj' ) ): modules.os.remove( 'fmiFunctions.obj' )
    # Compile FMU shared library.
    build_process = modules.subprocess.Popen( [ build_process_batch_file, fmi_model_identifier ] )
    stdout, stderr = build_process.communicate()

    if ( False == modules.os.path.isfile( fmu_shared_library_name ) ):
        raise RuntimeError( '\n[ERROR] Not able to create shared library: {}'.format( fmu_shared_library_name ) )

    return fmu_shared_library_name



# -----------------------------------------------------------------
# Copyright (c) 2022, AIT Austrian Institute of Technology GmbH.
# All rights reserved. See file FMIPP_LICENSE.txt for details.
# -----------------------------------------------------------------

#
# Collection of helper functions for creating FMU CS according to FMI 2.0
#

# Get templates for the XML model description depending on the FMI version.
def fmi2GetModelDescriptionTemplates( verbose, modules ):
    # Template string for XML model description.
    main = '<?xml version="1.0" encoding="UTF-8"?>\n<fmiModelDescription\n\txmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"\n\tfmiVersion="2.0"\n\tmodelName="__MODEL_NAME__"\n\tguid="{__GUID__}"\n\tgenerationTool="FMI++ TRNSYS Export Utility"\n\tauthor="__USER__"\n\tgenerationDateAndTime="__DATE_AND_TIME__"\n\tvariableNamingConvention="flat"\n\tnumberOfEventIndicators="0">\n\t<CoSimulation\n\t\tmodelIdentifier="__MODEL_IDENTIFIER__"\n\t\tneedsExecutionTool="true"\n\t\tcanHandleVariableCommunicationStepSize="true"\n\t\tcanNotUseMemoryManagementFunctions="true"\n\t\tcanInterpolateInputs="false"\n\t\tmaxOutputDerivativeOrder="0"\n\t\tcanGetAndSetFMUstate="false"\n\t\tprovidesDirectionalDerivative="false"/>\n\t<VendorAnnotations>\n\t\t<Tool name="FMI++Export">\n\t\t\t<Executable\n\t\t\t\texecutableURI="__PYTHON_URI__"\n\t\t\t\tentryPointURI="fmu://resources/"\n\t\t\t\targuments="run_backend___GUID__.py"\n\t\t\t\tpreArguments=""\n\t\t\t\tpostArguments=""/>__ADDITIONAL_FILES__</Tool>\n\t</VendorAnnotations>\n\t<ModelVariables>\n__MODEL_VARIABLES__\t</ModelVariables>\n\t<ModelStructure>\n\t\t<Outputs>\n__UNKNOWNS_OUTPUTS__\t\t</Outputs>\n\t\t<InitialUnknowns>\n__UNKNOWNS_INITIALS__\t\t</InitialUnknowns>\n\t</ModelStructure>\n</fmiModelDescription>'

    # Template string for XML model description of scalar variables.
    scalar_variable_node = '\t\t<ScalarVariable name="__VAR_NAME__" valueReference="__VAL_REF__" variability="__VARIABILITY__" causality="__CAUSALITY__" __INITIAL__>\n\t\t\t<__VAR_TYPE____START_VALUE__/>\n\t\t</ScalarVariable>\n'

    # Template string for XML model description of scalar unknown.
    scalar_unknown_node = '\t\t\t<Unknown index="__VAR_INDEX__"/>\n'

    return ( main, scalar_variable_node, scalar_unknown_node )


def fmi2addVariabilityAndCausalityToModelDescription( scalar_variable_description, type, is_input, is_parameter, verbose, modules ):
    if ( True == is_parameter ):
        scalar_variable_description = scalar_variable_description.replace( '__CAUSALITY__', 'parameter' )
        scalar_variable_description = scalar_variable_description.replace( '__VARIABILITY__', 'tunable' )
    elif ( True == is_input and False == is_parameter ):
        scalar_variable_description = scalar_variable_description.replace( '__CAUSALITY__', 'input' )
        if ( 'Real' == type ):
            scalar_variable_description = scalar_variable_description.replace( '__VARIABILITY__', 'continuous' )
        else:
            scalar_variable_description = scalar_variable_description.replace( '__VARIABILITY__', 'discrete' )
    elif ( False == is_input and False == is_parameter ):
        scalar_variable_description = scalar_variable_description.replace( '__CAUSALITY__', 'output' )
        if ( 'Real' == type ):
            scalar_variable_description = scalar_variable_description.replace( '__VARIABILITY__', 'continuous' )
        else:
            scalar_variable_description = scalar_variable_description.replace( '__VARIABILITY__', 'discrete' )
    return scalar_variable_description


# Add optional files to XML model description.
def fmi2AddOptionalFilesToModelDescription( optional_files, model_description, verbose, modules ):
    if ( 0 == len( optional_files ) ):
        model_description = model_description.replace( '__ADDITIONAL_FILES__', '' )
    else:
        additional_files_description = ''
        indent = '\n\t\t'

        for file_name in optional_files:
            additional_files_description += indent + '\t<File file=\"fmu://resources/' + modules.os.path.basename( file_name ) + '\"/>'
            if ( True == verbose ): modules.log( '[DEBUG] Added additional file to model description: ', modules.os.path.basename( file_name ) )
        additional_files_description += indent

        model_description = model_description.replace( '__ADDITIONAL_FILES__', additional_files_description )

    return model_description


# Create shared library for FMU.
def fmi2CreateSharedLibrary( fmi_model_identifier, verbose, modules ):
    # Check platform to determine shared library file extension.
    system = modules.platform.system()
    if system =='Linux':
      file_ending = '.so'
    elif system == 'Windows':
      file_ending = '.dll'
    else: # system == 'Darwin'
      file_ending = '.dylib'

    # Define name of shared library based on platform.
    fmu_shared_library_name = fmi_model_identifier + file_ending

    for file in modules.os.listdir(modules.os.path.join(modules.os.path.dirname(__file__), 'bin')):
      if ((file.startswith('libfmi2.') or file.startswith('fmi2.')) and file.endswith(file_ending)):
        fmi2_dll_path = modules.os.path.join( modules.os.path.dirname( __file__ ), 'bin', file)

    if ( False == modules.os.path.isfile( fmi2_dll_path ) ):
        raise RuntimeError( '\n[ERROR] DLL not found: {}'.format( fmi2_dll_path ) )
    modules.shutil.copy( fmi2_dll_path, fmu_shared_library_name )

    if ( False == modules.os.path.isfile( fmu_shared_library_name ) ):
        raise RuntimeError( '\n[ERROR] Not able to create shared library: {}'.format( fmu_shared_library_name ) )

    return fmu_shared_library_name



# -----------------------------------------------------------------
# Copyright (c) 2022, AIT Austrian Institute of Technology GmbH.
# All rights reserved. See file FMIPP_LICENSE.txt for details.
# -----------------------------------------------------------------

import fmipp
import sys

def runFMUBackend( backend_class_file_path, backend_class_name ):
	"""
	Initialize and run the FMU back-end.

	:param backend_class_file_path: path to file implementing the FMU back-end
	:param backend_class_name: name of class implementing the FMU back-end
	"""
	try:
		import importlib

		# Creating a ModuleSpec instance based on the path to the class file.
		spec = importlib.util.spec_from_file_location( backend_class_name, backend_class_file_path )

		# Load class inside separate module.
		module = importlib.util.module_from_spec( spec )
		spec.loader.exec_module( module )
	except:
		import imp

		module = imp.load_source( backend_class_name, backend_class_file_path )

	# Retrieve class implementing the FMU back-end.
	backend_class = getattr( module, backend_class_name )

	# Instantiate FMU back-end.
	backend = backend_class()

	try:
		# Initialize the FMU back-end.
		backend._initBackEnd()

		# Enter the simulation loop.
		backend._run()
	except Exception as ex:
		print( ex )



# -----------------------------------------------------------------
# Copyright (c) 2022, AIT Austrian Institute of Technology GmbH.
# All rights reserved. See file FMIPP_LICENSE.txt for details.
# -----------------------------------------------------------------

try:

    import numpy as np

    def _get_item( a, i ):
        return double_array_getitem( a, i )

    def getDerivatives( fmu_me ):
        '''
        Extract the derivatives matrix (as numpy.array) from an FMU for ModelExchange (FMI v1.0 or FMI 2.0).
        '''

        if not isinstance( fmu_me, FMUModelExchangeV1 ) and not isinstance( fmu_me, FMUModelExchangeV2 ):
            raise TypeError( 'this function is only supported for FMUModelExchangeV1 and FMUModelExchangeV2' )

        n = fmu_me.nStates()
        der = new_double_array( n )
        fmu_me.getDerivatives( der )

        return np.array( [ _get_item(der, i) for i in range(0, n) ] )

    def getJacobian( fmu2_me ):
        '''
        Extract the Jacobian matrix (as numpy.array) from an FMU for ModelExchange (FMI v2.0).
        '''

        if not isinstance( fmu2_me, FMUModelExchangeV2 ):
            raise TypeError( 'this function is only supported for FMUModelExchangeV2' )

        n = fmu2_me.nStates()
        jac = new_double_array( n * n )
        fmu2_me.getJac( jac )

        return np.array( [ [ _get_item(jac, i) for i in range(j*n, (j+1)*n)] for j in range(0, n) ] )

except ModuleNotFoundError:

    print( 'Unable to load fmipp.numeric. Probably numpy is not installed?' )



# -----------------------------------------------------------------
# Copyright (c) 2022, AIT Austrian Institute of Technology GmbH.
# All rights reserved. See file FMIPP_LICENSE.txt for details.
# -----------------------------------------------------------------

import warnings
import fmipp.export.fmippex as fmippex
import sys

import abc
try:
	import abc.ABC as ABC
except:
	pass

try:
	from abc import ABCMeta
	class ABC:
		__metaclass__ = ABCMeta
except:
	pass


class FMIAdapterBase( ABC ):
	"""
	Base class for FMI adapter implementations.
	"""

	@abc.abstractmethod
	def init( self, currentCommunicationPoint ):
		"""
		Initialize the FMU (definition of input/output variables and parameters, enforce step size).
		:param currentCommunicationPoint: current communication point of the master algorithm during initialization (float)
		"""
		pass


	@abc.abstractmethod
	def doStep( self, currentCommunicationPoint, communicationStepSize ):
		"""
		Make a simulation step.
		:param currentCommunicationPoint: current communication point of the master algorithm (float)
		:param communicationStepSize: communication step size (float)
		"""
		pass


	@abc.abstractmethod
	def defineRealParameters( self, *parameterNames ):
		"""
		Define parameters (of type real).
		:param *parameterNames: names of parameters of type real
		"""
		pass


	@abc.abstractmethod
	def defineIntegerParameters( self, *parameterNames ):
		"""
		Define parameters (of type integer).
		:param *parameterNames: names of parameters of type integer
		"""
		pass


	@abc.abstractmethod
	def defineBooleanParameters( self, *parameterNames ):
		"""
		Define parameters (of type boolean).
		:param *parameterNames: names of parameters of type boolean
		"""
		pass


	@abc.abstractmethod
	def defineStringParameters( self, *parameterNames ):
		"""
		Define parameters (of type string).
		:param *parameterNames: names of parameters of type string
		"""
		pass


	@abc.abstractmethod
	def defineRealInputs( self, *inputVariableNames ):
		"""
		Define input variables (of type real).
		:param *inputVariableNames: names of input variables of type real
		"""
		pass


	@abc.abstractmethod
	def defineIntegerInputs( self, *inputVariableNames ):
		"""
		Define input variables (of type integer).
		:param *inputVariableNames: names of input variables of type integer
		"""
		pass


	@abc.abstractmethod
	def defineBooleanInputs( self, *inputVariableNames ):
		"""
		Define input variables (of type boolean).
		:param *inputVariableNames: names of input variables of type boolean
		"""
		pass


	@abc.abstractmethod
	def defineStringInputs( self, *inputVariableNames ):
		"""
		Define input variables (of type string).
		:param *inputVariableNames: names of input variables of type string
		"""
		pass


	@abc.abstractmethod
	def defineRealOutputs( self, *outputVariableNames ):
		"""
		Define output variables (of type real).
		:param *outputVariableNames: names of output variables of type real
		"""
		pass


	@abc.abstractmethod
	def defineIntegerOutputs( self, *outputVariableNames ):
		"""
		Define output variables (of type integer).
		:param *outputVariableNames: names of output variables of type integer
		"""
		pass


	@abc.abstractmethod
	def defineBooleanOutputs( self, *outputVariableNames ):
		"""
		Define output variables (of type boolean).
		:param *outputVariableNames: names of output variables of type boolean
		"""
		pass


	@abc.abstractmethod
	def defineStringOutputs( self, *outputVariableNames ):
		"""
		Define output variables (of type string).
		:param *outputVariableNames: names of output variables of type string
		"""
		pass


	@abc.abstractmethod
	def getRealParameterValues( self ):
		"""
		Retrieve parameters (of type real).
		:return: dict of parameters of type real (includes all parameters of type real previously defined during the initialization)
		"""
		pass


	@abc.abstractmethod
	def getIntegerParameterValues( self ):
		"""
		Retrieve parameters (of type integer).
		:return: dict of parameters of type integer (includes all parameters of type integer previously defined during the initialization)
		"""
		pass


	@abc.abstractmethod
	def getBooleanParameterValues( self ):
		"""
		Retrieve parameters (of type boolean).
		:return: dict of parameters of type boolean (includes all parameters of type boolean previously defined during the initialization)
		"""
		pass


	@abc.abstractmethod
	def getStringParameterValues( self ):
		"""
		Retrieve parameters (of type string).
		:return: dict of parameters of type string (includes all parameters of type string previously defined during the initialization)
		"""
		pass


	@abc.abstractmethod
	def getRealInputValues( self ):
		"""
		Retrieve input variables (of type real).
		:return: dict of input variables of type real (includes all input variables of type real previously defined during the initialization)
		"""
		pass


	@abc.abstractmethod
	def getIntegerInputValues( self ):
		"""
		Retrieve input variables (of type integer).
		:return: dict of input variables of type real (includes all input variables of type real previously defined during the initialization)
		"""
		pass


	@abc.abstractmethod
	def getBooleanInputValues( self ):
		"""
		Retrieve input variables (of type boolean).
		:return: dict of input variables of type boolean (includes all input variables of type boolean previously defined during the initialization)
		"""
		pass


	@abc.abstractmethod
	def getStringInputValues( self ):
		"""
		Retrieve input variables (of type string).
		:return: dict of input variables of type string (includes all input variables of type string previously defined during the initialization)
		"""
		pass


	@abc.abstractmethod
	def setRealOutputValues( self, outputValues ):
		"""
		Set output variables (of type real).
		:param outputValues: dict of output variables of type real (must include all output variables of type real previously defined during the initialization)
		"""
		pass


	@abc.abstractmethod
	def setIntegerOutputValues( self, outputValues ):
		"""
		Set output variables (of type integer).
		:param outputValues: dict of output variables of type integer (must include all output variables of type integer previously defined during the initialization)
		"""
		pass


	@abc.abstractmethod
	def setBooleanOutputValues( self, outputValues ):
		"""
		Set output variables (of type boolean).
		:param outputValues: dict of output variables of type boolean (must include all output variables of type boolean previously defined during the initialization)
		"""
		pass


	@abc.abstractmethod
	def setStringOutputValues( self, outputValues ):
		"""
		Set output variables (of type string).
		:param outputValues: dict of output variables of type string (must include all output variables of type string previously defined during the initialization)
		"""
		pass


	def __init__( self ):
		""" Constructor."""
		# This flag indicates whether the FMI export interface is activated.
		self._fmippexActive = False

		# FMI component backend.
		self._backend = None

		# Flag to indicate that a fixed step size is to be enforced.
		self._enforceTimeStep = False

		# Enforced step size.
		self._enforcedTimeStepSize = float( 'NAN' )

		# Real parameters.
		self._realParameters = None
		self._realParameterSize = 0
		self._realParameterNames = []
		self._debugRealParameterValues = []

		# Real input variables.
		self._realInputs = None
		self._realInputSize = 0
		self._realInputNames = []
		self._debugRealInputValues = []

		# Real output variables.
		self._realOutputs = None
		self._realOutputSize = 0
		self._realOutputNames = []
		self._debugRealOutputValues = []

		# Integer parameters.
		self._integerParameters = None
		self._integerParameterSize = 0
		self._integerParameterNames = []
		self._debugIntegerParameterValues = []

		# Integer input variables.
		self._integerInputs = None
		self._integerInputSize = 0
		self._integerInputNames = []
		self._debugIntegerInputValues = []

		# Integer output variables.
		self._integerOutputs = None
		self._integerOutputSize = 0
		self._integerOutputNames = []
		self._debugIntegerOutputValues = []

		# Boolean parameters.
		self._booleanParameters = None
		self._booleanParameterSize = 0
		self._booleanParameterNames = []
		self._debugBooleanParameterValues = []

		# Boolean input variables.
		self._booleanInputs = None
		self._booleanInputSize = 0
		self._booleanInputNames = []
		self._debugBooleanInputValues = []

		# Boolean output variables.
		self._booleanOutputs = None
		self._booleanOutputSize = 0
		self._booleanOutputNames = []
		self._debugBooleanOutputValues = []

		# String parameters.
		self._stringParameters = None
		self._stringParameterSize = 0
		self._stringParameterNames = []
		self._debugStringParameterValues = []

		# String input variables.
		self._stringInputs = None
		self._stringInputSize = 0
		self._stringInputNames = []
		self._debugStringInputValues = []

		# String output variables.
		self._stringOutputs = None
		self._stringOutputSize = 0
		self._stringOutputNames = []
		self._debugStringOutputValues = []


	def enforceTimeStep( self, stepSize ):
		"""
		Enforce time step, call from init(...) function.
		:param stepSize: value of the enforced step size (float)
		"""
		self._enforceTimeStep = True
		self._enforcedTimeStepSize = stepSize


	def checkEnforceTimeStep( self ):
		"""Check if fixed time steps are enforced."""
		return self._enforceTimeStep


	def debugSetRealParameterValues( self, realParameterValues ):
		"""
		Set parameters of type real (in debug mode).
		:param realParameterValues: dict of parameters of type real (must include all parameters of type real previously defined during the initialization)
		"""
		# Sanity check: all defined parameters have to be given as input.
		for name in self._realParameterNames:
			if not name in realParameterValues:
				raise RuntimeError( '[FMIAdapter:debugSetRealParameterValues] parameter \'{}\' is missing.'.format( name ) )

		for i in range( 0, self._realParameterSize ):
			self._debugRealParameterValues[i] = realParameterValues[ self._realParameterNames[i] ]


	def debugSetRealInputValues( self, realInputValues ):
		"""
		Set input values of type real (in debug mode).
		:param realInputValues: dict of input variables of type real (must include all input values of type real previously defined during the initialization)
		"""
		# Sanity check: all defined input variables have to be given as input.
		for name in self._realInputNames:
			if not name in realInputValues:
				raise RuntimeError( '[FMIAdapter:debugSetRealInputValues] input variable \'{}\' is missing.'.format( name ) )

		for i in range( 0, self._realInputSize ):
			self._debugRealInputValues[i] = realInputValues[ self._realInputNames[i] ]


	def debugGetRealOutputValues( self ):
		"""
		Get output values of type real (in debug mode).
		:return: dict of output variables of type real (includes all output variables of type real previously defined during the initialization)
		"""
		debugRealOutputs = {}

		for i in range( 0, self._realOutputSize ):
			debugRealOutputs[ self._realOutputNames[i] ] = self._debugRealOutputValues[i]

		return debugRealOutputs


	def debugSetIntegerParameterValues( self, integerParameterValues ):
		"""
		Set parameters of type integer (in debug mode).
		:param integerParameterValues: dict of parameters of type integer (must include all parameters of type integer previously defined during the initialization)
		"""
		# Sanity check: all defined parameters have to be given as input.
		for name in self._integerParameterNames:
			if not name in integerParameterValues:
				raise RuntimeError( '[FMIAdapter:debugSetIntegerParameterValues] parameter \'{}\' is missing.'.format( name ) )

		for i in range( 0, self._integerParameterSize ):
			self._debugIntegerParameterValues[i] = integerParameterValues[ self._integerParameterNames[i] ]


	def debugSetIntegerInputValues( self, integerInputValues ):
		"""
		Set input values of type integer (in debug mode).
		:param integerInputValues: dict of input variables of type integer (must include all input variables of type integer previously defined during the initialization)
		"""
		# Sanity check: all defined input variables have to be given as input.
		for name in self._integerInputNames:
			if not name in integerInputValues:
				raise RuntimeError( '[FMIAdapter:debugSetIntegerInputValues] input variable \'{}\' is missing.'.format( name ) )

		for i in range( 0, self._integerInputSize ):
			self._debugIntegerInputValues[i] = integerInputValues[ self._integerInputNames[i] ]


	def debugGetIntegerOutputValues( self ):
		"""
		Get output values of type integer (in debug mode).
		:return: dict of output variables of type integer (includes all output variables of type integer previously defined during the initialization)
		"""
		debugIntegerOutputs = {}

		for i in range( 0, self._integerOutputSize ):
			debugIntegerOutputs[ self._integerOutputNames[i] ] = self._debugIntegerOutputValues[i]

		return debugIntegerOutputs


	def debugSetBooleanParameterValues( self, booleanParameterValues ):
		"""
		Set parameters of type boolean (in debug mode).
		:param booleanParameterValues: dict of parameters of type boolean (must include all parameters of type boolean previously defined during the initialization)
		"""
		# Sanity check: all defined parameters have to be given as input.
		for name in self._booleanParameterNames:
			if not name in booleanParameterValues:
				raise RuntimeError( '[FMIAdapter:debugSetBooleanParameterValues] parameter \'{}\' is missing.'.format( name ) )

		for i in range( 0, self._booleanParameterSize ):
			self._debugBooleanParameterValues[i] = booleanParameterValues[ self._booleanParameterNames[i] ]


	def debugSetBooleanInputValues( self, booleanInputValues ):
		"""
		Set input values of type boolean (in debug mode).
		:param booleanInputValues: dict of input variables of type boolean (must include all input variables of type boolean previously defined during the initialization)
		"""
		# Sanity check: all defined input variables have to be given as input.
		for name in self._booleanInputNames:
			if not name in booleanInputValues:
				raise RuntimeError( '[FMIAdapter:debugSetBooleanInputValues] input variable \'{}\' is missing.'.format( name ) )

		for i in range( 0, self._booleanInputSize ):
			self._debugBooleanInputValues[i] = booleanInputValues[ self._booleanInputNames[i] ]


	def debugGetBooleanOutputValues( self ):
		"""
		Get output values of type boolean (in debug mode).
		:return: dict of output variables of type boolean (includes all output variables of type boolean previously defined during the initialization)
		"""
		debugBooleanOutputs = {}

		for i in range( 0, self._booleanOutputSize ):
			debugBooleanOutputs[ self._booleanOutputNames[i] ] = self._debugBooleanOutputValues[i]

		return debugBooleanOutputs


	def debugSetStringParameterValues( self, stringParameterValues ):
		"""
		Set parameters of type string (in debug mode).
		:param stringParameterValues: dict of parameters of type string (must include all parameters of type string previously defined during the initialization)
		"""
		# Sanity check: all defined parameters have to be given as input.
		for name in self._stringParameterNames:
			if not name in stringParameterValues:
				raise RuntimeError( '[FMIAdapter:debugSetStringParameterValues] parameter \'{}\' is missing.'.format( name ) )

		for i in range( 0, self._stringParameterSize ):
			self._debugStringParameterValues[i] = stringParameterValues[ self._stringParameterNames[i] ]


	def debugSetStringInputValues( self, stringInputValues ):
		"""
		Set input values of type string (in debug mode).
		:param stringInputValues: dict of input variables of type string (must include all input variables of type string previously defined during the initialization)
		"""
		# Sanity check: all defined input variables have to be given as input.
		for name in self._stringInputNames:
			if not name in stringInputValues:
				raise RuntimeError( '[FMIAdapter:debugSetStringInputValues] input variable \'{}\' is missing.'.format( name ) )

		for i in range( 0, self._stringInputSize ):
			self._debugStringInputValues[i] = stringInputValues[ self._stringInputNames[i] ]


	def debugGetStringOutputValues( self ):
		"""
		Get output values of type string (in debug mode).
		:return: dict of output variables of type string (includes all output variables of type string previously defined during the initialization)
		"""
		debugStringOutputs = {}

		for i in range( 0, self._stringOutputSize ):
			debugStringOutputs[ self._stringOutputNames[i] ] = self._debugStringOutputValues[i]

		return debugStringOutputs


	def _defineVariables( self, names, size, variables, type, debugValues, funcInitParams ):
		"""Generic function for the definition of variables (inputs, outputs, parameters)"""
		# Check if export interface is active.
		if self._fmippexActive is False:
			warnings.warn( 'FMI++ export interface is not active.', RuntimeWarning )
			try:
				debugValues.clear()
			except:
				while len(debugValues) > 0 : debugValues.pop()
			debugValues.extend( [ type() ] * size )
			return fmippex.fmi2OK

		# Define variable names.
		labels = fmippex.new_string_array( size )
		for i in range( 0, size ): fmippex.string_array_setitem( labels, i, names[i] )

		# Initialize variables.
		return funcInitParams( labels, variables, size )


	def _getValues( self, names, size, variables, type, debugValues, funcGetValues, funcGetItem ):
		"""Generic getter function (inputs, parameters)"""
		# Check if export interface is active.
		if self._fmippexActive is False:
			warnings.warn( 'FMI++ export interface is not active.', RuntimeWarning )
			return dict( [ ( n, v ) for n, v in zip( names, debugValues ) ] )

		# Read parameters.
		status = funcGetValues( variables, size )
		if not status is fmippex.fmi2OK: return None

		getValues = [ type() ] * size
		for i in range( 0, size ): getValues[i] = type( funcGetItem( variables, i ) )
		return dict( [ ( n, v ) for n, v in zip( names, getValues ) ] )


	def _setValues( self, names, size, variables, values, debugValues, funcSetValues, funcSetItem ):
		"""Generic setter function (outputs)"""
		# Check if export interface is active.
		if self._fmippexActive is False:
			warnings.warn( 'FMI++ export interface is not active.', RuntimeWarning )
			try:
				debugValues.clear()
			except:
				while len(debugValues) > 0 : debugValues.pop()
			debugValues.extend( [ values[ name ] for name in names ] )
			return fmippex.fmi2OK

		# Set output values.
		for i in range( 0, size ): funcSetItem( variables, i, values[ names[i] ] )

		# Write current outputs.
		return funcSetValues( variables, size )


	def _initBackEnd( self ):
		"""Initialize base. Call this function in order to activate the FMI export interface during co-simulation."""
		# Variable that indicates if FMI++ export interface is active.
		self._fmippexActive = True

		# Create a new FMI backend.
		self._backend = fmippex.FMIComponentBackEnd()

		# Start the initialization of the backend.
		initStatus = self._backend.startInitialization()
		if not initStatus is fmippex.fmi2OK:
			raise RuntimeError( '[FMIAdapter:initBackEnd] start of initialization of FMI++ interface unsuccessful' )

		self.init( self._backend.getCurrentCommunicationPoint() )

		if self._enforceTimeStep is True:
			# Let's do fixed time steps!
			self._backend.enforceTimeStep( self._enforcedTimeStepSize )

		# End the initialization of the backend.
		initStatus = self._backend.endInitialization()

		if not initStatus is fmippex.fmi2OK:
			raise RuntimeError( '[FMIAdapter:initBackEnd] end of initialization of FMI++ interface unsuccessful' )


	def _run( self ):
		"""Iterate the FMU."""
		while self._fmippexActive is True:
			# Wait for simulation master to hand over control.
			self._backend.waitForMaster()

			# Make a step.
			self.doStep( self._backend.getCurrentCommunicationPoint(), self._backend.getCommunicationStepSize() )

			if self._enforceTimeStep is True:
				# Let's do fixed time steps!
				self._backend.enforceTimeStep( self._enforcedTimeStepSize )

			# Give back control to simulation master.
			self._backend.signalToMaster()
